#define WRITE_ON_STACK

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <sys/syscall.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>

#if defined __x86_64__
#   define RAX rax
#   define RBX rbx
#   define RCX rcx
#   define RDX rdx
#   define RDI rdi
#   define RSI rsi
#   define RSP rsp
#   define RIP rip
#else
#   define RAX eax
#   define RBX ebx
#   define RCX ecx
#   define RDX edx
#   define RDI edi
#   define RSI esi
#   define RSP esp
#   define RIP eip
#endif

#if defined __x86_64__
#   define SYSCALL 0x050fL
#else
#   define SYSCALL 0x80cd
#endif

int main(int argc, char *argv[])
{
    struct user_regs_struct oldregs, regs;
    long int inst = 0;
#if defined WRITE_ON_STACK
    long int stack = 0;
#endif
    pid_t pid = atoi(argv[1]);

    fprintf(stderr, "attaching... ");

    /* Attach process */
    if(ptrace(PTRACE_ATTACH, pid, 0, 0) < 0)
    {
        perror("ptrace_attach");
        return errno;
    }
    waitpid(pid, NULL, 0);

    /* Wait until the attached process uses a system call */
    while(inst != SYSCALL)
    {
        if(ptrace(PTRACE_SYSCALL, pid, NULL, 0) < 0)
        {
            perror("ptrace_syscall");
            return errno;
        }
        waitpid(pid, NULL, 0);

        if(ptrace(PTRACE_GETREGS, pid, NULL, &oldregs) < 0)
        {
            perror("ptrace_getregs");
            return errno;
        }

        inst = ptrace(PTRACE_PEEKTEXT, pid, oldregs.RIP - 2, 0) & 0xffff;
    }

    fprintf(stderr, "found syscall, injecting code... ");

    /* Prepare kill(0, 9) call */
    regs = oldregs;

#if defined WRITE_ON_STACK
    /* Write the syscall instruction to the stack and set the instruction
     * pointer to that address */
    stack = ptrace(PTRACE_PEEKTEXT, pid, oldregs.RSP - sizeof(long), 0);
    ptrace(PTRACE_POKETEXT, pid, oldregs.RSP - sizeof(long), SYSCALL);
    regs.RIP = oldregs.RSP - sizeof(long);
#else
    /* Rewind 2 bytes before the instruction pointer: the syscall instruction
     * is there */
    regs.RIP = regs.RIP - 2;
#endif

    regs.RAX = SYS_kill;
#if defined __x86_64__
    regs.RDI = 0;
    regs.RSI = 9;
#else
    regs.RBX = 0;
    regs.RCX = 9;
#endif

    if(ptrace(PTRACE_SETREGS, pid, NULL, &regs) < 0)
    {
        perror("ptrace_setregs (1)");
        return errno;
    }

    /* Call kill(0, 9) */
    if(ptrace(PTRACE_SINGLESTEP, pid, NULL, NULL) < 0)
    {
        perror("ptrace_singlestep");
        return errno;
    }
    waitpid(pid, NULL, 0);

    /* Restore registers */
    if(ptrace(PTRACE_SETREGS, pid, NULL, &oldregs) < 0)
    {
        perror("ptrace_setregs(2)");
        return errno;
    }

#if defined WRITE_ON_STACK
    /* Restore the stack contents (probably not needed since we wrote before
     * the stack pointer, but we shouldn't assume anything) */
    if(ptrace(PTRACE_POKETEXT, pid, oldregs.RSP - sizeof(long), stack) < 0)
    {
        perror("ptrace_pokedata(2)");
        return errno;
    }
#endif

    /* Resume execution */
    ptrace(PTRACE_DETACH, pid, 0, 0);

    fprintf(stderr, "done.\n");

    return 0;
}

